//
// Copyright (c) 2015 The ANGLE Project Authors. All rights reserved.
// Use of this source code is governed by a BSD-style license that can be
// found in the LICENSE file.
//

// FunctionsWGL.h: Implements the FuntionsWGL class.

#include "libANGLE/renderer/gl/wgl/FunctionsWGL.h"

namespace rx
{

template <typename T>
static void GetWGLProcAddress(HMODULE glModule, PFNWGLGETPROCADDRESSPROC getProcAddressWGL,
                              const std::string &procName, T *outProcAddress)
{
    T proc = nullptr;
    if (getProcAddressWGL)
    {
        proc = reinterpret_cast<T>(getProcAddressWGL(procName.c_str()));
    }

    if (!proc)
    {
        proc = reinterpret_cast<T>(GetProcAddress(glModule, procName.c_str()));
    }

    *outProcAddress = proc;
}

template <typename T>
static void GetWGLExtensionProcAddress(HMODULE glModule, PFNWGLGETPROCADDRESSPROC getProcAddressWGL,
                                       const std::string &extensions, const std::string &extensionName,
                                       const std::string &procName, T *outProcAddress)
{
    T proc = nullptr;
    if (extensions.find(extensionName) != std::string::npos)
    {
        GetWGLProcAddress(glModule, getProcAddressWGL, procName, &proc);
    }

    *outProcAddress = proc;
}

FunctionsWGL::FunctionsWGL()
    : copyContext(nullptr),
      createContext(nullptr),
      createLayerContext(nullptr),
      deleteContext(nullptr),
      getCurrentContext(nullptr),
      getCurrentDC(nullptr),
      getProcAddress(nullptr),
      makeCurrent(nullptr),
      shareLists(nullptr),
      useFontBitmapsA(nullptr),
      useFontBitmapsW(nullptr),
      swapBuffers(nullptr),
      useFontOutlinesA(nullptr),
      useFontOutlinesW(nullptr),
      describeLayerPlane(nullptr),
      setLayerPaletteEntries(nullptr),
      getLayerPaletteEntries(nullptr),
      realizeLayerPalette(nullptr),
      swapLayerBuffers(nullptr),
      swapMultipleBuffers(nullptr),
      createContextAttribsARB(nullptr),
      getPixelFormatAttribivARB(nullptr),
      getExtensionStringEXT(nullptr),
      getExtensionStringARB(nullptr),
      swapIntervalEXT(nullptr),
      createPbufferARB(nullptr),
      getPbufferDCARB(nullptr),
      releasePbufferDCARB(nullptr),
      destroyPbufferARB(nullptr),
      queryPbufferARB(nullptr),
      bindTexImageARB(nullptr),
      releaseTexImageARB(nullptr),
      setPbufferAttribARB(nullptr)
{
}

void FunctionsWGL::initialize(HMODULE glModule, HDC context)
{
    // First grab the wglGetProcAddress function from the gl module
    GetWGLProcAddress(glModule, nullptr, "wglGetProcAddress", &getProcAddress);

    // Load the core wgl functions
    GetWGLProcAddress(glModule, getProcAddress, "wglCopyContext", &copyContext);
    GetWGLProcAddress(glModule, getProcAddress, "wglCreateContext", &createContext);
    GetWGLProcAddress(glModule, getProcAddress, "wglCreateLayerContext", &createLayerContext);
    GetWGLProcAddress(glModule, getProcAddress, "wglDeleteContext", &deleteContext);
    GetWGLProcAddress(glModule, getProcAddress, "wglGetCurrentContext", &getCurrentContext);
    GetWGLProcAddress(glModule, getProcAddress, "wglGetCurrentDC", &getCurrentDC);
    GetWGLProcAddress(glModule, getProcAddress, "wglMakeCurrent", &makeCurrent);
    GetWGLProcAddress(glModule, getProcAddress, "wglShareLists", &shareLists);
    GetWGLProcAddress(glModule, getProcAddress, "wglUseFontBitmapsA", &useFontBitmapsA);
    GetWGLProcAddress(glModule, getProcAddress, "wglUseFontBitmapsW", &useFontBitmapsW);
    swapBuffers = SwapBuffers; // SwapBuffers is statically linked from GDI
    GetWGLProcAddress(glModule, getProcAddress, "wglUseFontOutlinesA", &useFontOutlinesA);
    GetWGLProcAddress(glModule, getProcAddress, "wglUseFontOutlinesW", &useFontOutlinesW);
    GetWGLProcAddress(glModule, getProcAddress, "wglDescribeLayerPlane", &describeLayerPlane);
    GetWGLProcAddress(glModule, getProcAddress, "wglSetLayerPaletteEntries", &setLayerPaletteEntries);
    GetWGLProcAddress(glModule, getProcAddress, "wglGetLayerPaletteEntries", &getLayerPaletteEntries);
    GetWGLProcAddress(glModule, getProcAddress, "wglRealizeLayerPalette", &realizeLayerPalette);
    GetWGLProcAddress(glModule, getProcAddress, "wglSwapLayerBuffers", &swapLayerBuffers);
    GetWGLProcAddress(glModule, getProcAddress, "wglSwapMultipleBuffers", &swapMultipleBuffers);

    // Load extension string getter functions
    GetWGLProcAddress(glModule, getProcAddress, "wglGetExtensionsStringEXT", &getExtensionStringEXT);
    GetWGLProcAddress(glModule, getProcAddress, "wglGetExtensionsStringARB", &getExtensionStringARB);

    std::string extensions = "";
    if (getExtensionStringEXT)
    {
        extensions = getExtensionStringEXT();
    }
    else if (getExtensionStringARB && context)
    {
        extensions = getExtensionStringARB(context);
    }

    // Load the wgl extension functions by checking if the context supports the extension first
    GetWGLExtensionProcAddress(glModule, getProcAddress, extensions, "WGL_ARB_create_context", "wglCreateContextAttribsARB", &createContextAttribsARB);
    GetWGLExtensionProcAddress(glModule, getProcAddress, extensions, "WGL_ARB_pixel_format", "wglGetPixelFormatAttribivARB", &getPixelFormatAttribivARB);
    GetWGLExtensionProcAddress(glModule, getProcAddress, extensions, "WGL_EXT_swap_control", "wglSwapIntervalEXT", &swapIntervalEXT);

    // WGL_ARB_pbuffer
    GetWGLExtensionProcAddress(glModule, getProcAddress, extensions, "WGL_ARB_pbuffer", "wglCreatePbufferARB", &createPbufferARB);
    GetWGLExtensionProcAddress(glModule, getProcAddress, extensions, "WGL_ARB_pbuffer", "wglGetPbufferDCARB", &getPbufferDCARB);
    GetWGLExtensionProcAddress(glModule, getProcAddress, extensions, "WGL_ARB_pbuffer", "wglReleasePbufferDCARB", &releasePbufferDCARB);
    GetWGLExtensionProcAddress(glModule, getProcAddress, extensions, "WGL_ARB_pbuffer", "wglDestroyPbufferARB", &destroyPbufferARB);
    GetWGLExtensionProcAddress(glModule, getProcAddress, extensions, "WGL_ARB_pbuffer", "wglQueryPbufferARB", &queryPbufferARB);

    // WGL_ARB_render_texture
    GetWGLExtensionProcAddress(glModule, getProcAddress, extensions, "WGL_ARB_render_texture", "wglBindTexImageARB", &bindTexImageARB);
    GetWGLExtensionProcAddress(glModule, getProcAddress, extensions, "WGL_ARB_render_texture", "wglReleaseTexImageARB", &releaseTexImageARB);
    GetWGLExtensionProcAddress(glModule, getProcAddress, extensions, "WGL_ARB_render_texture", "wglSetPbufferAttribARB", &setPbufferAttribARB);
}

}
